#include <curses.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <list>
#include <vector>
#include <map>
#include <assert.h>
using namespace std;

#include "dissolve.h"
#include "random.h"
#include "view.h"
#include "shapes.h"

const int CLIFF_MAX = 50;
const int CLIFF_MIN = 35;

const int LAVA_POOL_DEPTH = 20;

struct dissolving_cavern
{
    int w,h;

    int drip_rate;
    unsigned short init_height;
    //int init_hardness_variation;
    int dissolve_vert;
    int gravity;
    int hard_areas;
    int encase;

    static const int lat_down = 1;
    static const int max_vel = 20;

    Array<int> height;
    Array<int> numdrips;

    list<dripper>& drippers;

    dissolving_cavern(int w, int h, list<dripper>& drippers_in) :
	w(w), h(h), drippers(drippers_in)
    {
	drip_rate = 20;
	init_height = max(3, int(20 + 5*gaussian()));
	absMinHeight = -h-w/2;
	dissolve_vert = 65 + rani(20);
	gravity = 17+rani(7);
	hard_areas = int((w*h/500)+(w*h/1000)*gaussian());
	encase = 4 + rani(5);

	//gravity = one_chance_in(10)? 1 : 3+rani(8);
	//init_hardness_variation = (init_height*(1+rani(9)))/10;

	height.init(w,h, init_height);
	numdrips.init(w,h, 0);
    }

    int absMinHeight;
    int getMinHeight(Coord c) {
	for (list<dripper>::iterator dripper = drippers.begin();
		dripper != drippers.end(); dripper++) {
	    const int dsq = (c - dripper->source).sqLen();
	    if (dsq < 25)
		return -1-dsq;
	}
	return absMinHeight;
	//return max(absMinHeight,
		//(-h/2 + abs(c.y - h/2) - c.x)*2);
    }

    void make_rock();
    void dissolve();

    //void writeDebug(int time);
    int debugView();

    bool in_bounds(Coord c) const {
	return (0 <= c.x && c.x <= w-1 &&
		0 <= c.y && c.y <= h-1);
    }
    bool in_bounds_strict(Coord c) const {
	return (0 < c.x && c.x < w-1 &&
		0 < c.y && c.y < h-1);
    }
    bool on_bounds(Coord c) const {
	return (0 == c.x || c.x == w-1 ||
		0 == c.y || c.y == h-1);
    }
};
DissolvedCavern::DissolvedCavern(CavernData data) : Cavern(data) {
    source = Coord(w/2, h/2);
    sourceDir = orthDirs[rani(4)];
    do {
	map.init(w, h, TILE_NULL);

	list<dripper> drippers;
	drippers.push_back(dripper(source, sourceDir*10));
	dissolving_cavern cavern(w, h, drippers);

	cavern.make_rock();
	cavern.dissolve();

	setTile(cavern, source);

	floodLava(cavern, lavaPos);
    } while (!acceptable());
}
bool DissolvedCavern::acceptable() {
    // Check that the lava is not too easy or too hard to get to, using
    // Dijkstra pathing
    static const int MIN_LAVA_DIST = 200;
    static const int MAX_LAVA_DIST = 300;
    std::map<Coord,int> unvisited;
    set<Coord> visited;
    unvisited[source] = 0;

    std::map<Coord,int>::iterator i, least;
    Coord c,nc;
    int d,nd,dir;
    while (true) {
	if (unvisited.empty()) {
#ifdef DEBUG
	    mvaddstr(25, 0, "No path to lava!");
	    clrtoeol();
	    refresh();
	    sleep(1);
#endif
	    return false;
	}
	d = MAX_LAVA_DIST;
	for (i = unvisited.begin(); i != unvisited.end(); i++)
	    if (i->second < d) {
		d = i->second;
		least = i;
	    }
	if (d == MAX_LAVA_DIST) {
#ifdef DEBUG
	    mvaddstr(25, 0, "Lava too far away!");
	    clrtoeol();
	    refresh();
	    sleep(1);
#endif
	    return false;
	}
	c = least->first;
	unvisited.erase(least);

	if (map(c) == TILE_LAVA) {
#ifdef DEBUG
	    char buf[80];
	    snprintf(buf, 80, "lava distance: %d", d);
	    mvaddstr(25, 0, buf);
	    clrtoeol();
	    refresh();
	    sleep(1);
#endif
	    return (d >= MIN_LAVA_DIST);
	}

	for (dir=0; dir<8; dir++) {
	    nc = c + dirs[dir];
	    if (visited.find(nc) == visited.end()) {
		i = unvisited.find(nc);
		if ( map.in_bounds(nc) && map(nc) != TILE_WALL &&
			!(map(c) == TILE_LOW && map(nc) == TILE_HIGH) ) {
		    nd = d + ((map(c) == TILE_HIGH && map(nc) == TILE_LOW) ?
			    50 : 1);
		    if (i == unvisited.end() || nd < i->second)
			unvisited[nc]=nd;
		}
	    }
	}
	visited.insert(c);
    }
}
void DissolvedCavern::setTile(const dissolving_cavern& cavern, Coord c) {
    set<Coord> toSet, toTrace;
    toSet.insert(c);
    while (!toSet.empty()) {
	c = *toSet.begin();
	toSet.erase(toSet.begin());

	int ht = cavern.height(c);

	if( ht > 0 || cavern.on_bounds(c)) {
	    map(c) = TILE_WALL;
	    continue;
	}

	if( ht <= cavern.absMinHeight )
	    lavaPos = c;

	Coord fromSource = c-source;
	if (fromSource == Coord(0,0))
	    map(c) = TILE_EXIT;
	else if (fromSource.lenDiamond() < 2*cavern.encase+4
		&& fromSource.length() < 2*cavern.encase+gaussian())
	    map(c) = TILE_CAVE_ENTRANCE;
	else
	    map(c) = Tile(TILE_FLOOR1 +
		((TILE_FLOOR_END-TILE_FLOOR1)*ht)/cavern.absMinHeight);

	for (int d=0; d<8; d++) {
	    Coord dir = dirs[d];
	    if (cavern.in_bounds(c+dir)) {
		if (map(c+dir) == TILE_NULL)
		    toSet.insert(c+dir);
		if (floorTile(map(c)) && floorTile(map(c+dir)) &&
			cavern.height(c+dir) <= ht - CLIFF_MAX) {
		    toTrace.insert(c);
		}
	    }
	}
    }
    for (set<Coord>::iterator i = toTrace.begin(); i != toTrace.end(); i++)
	traceEdge(cavern, *i);
}

void DissolvedCavern::traceEdge(const dissolving_cavern& cavern, Coord c) {
    int ht = cavern.height(c);
    for (int d=0; d<8; d++) {
	Coord dir = dirs[d];
	if (cavern.in_bounds(c+dir)) {
	    if (floorTile(map(c+dir))) {
		if (cavern.height(c+dir) <= ht - CLIFF_MIN) {
		    map(c) = TILE_HIGH;
		    map(c+dir) = TILE_LOW;
		    traceEdge(cavern, c+dir);
		}
		else if (cavern.height(c+dir) >= ht + CLIFF_MIN) {
		    map(c) = TILE_LOW;
		    map(c+dir) = TILE_HIGH;
		    traceEdge(cavern, c+dir);
		}
	    }
	}
    }
}

void DissolvedCavern::floodLava(const dissolving_cavern& cavern, Coord c) {
    map(c) = TILE_LAVA;
    for (int d=0; d<8; d++) {
	Coord c2 = c+dirs[d];
	if (cavern.in_bounds(c2) && floorTile(map(c2)) &&
		cavern.height(c2) <= cavern.absMinHeight + LAVA_POOL_DEPTH)
	    floodLava(cavern, c2);
    }
}

Tile DissolvedCavern::getTile(Coord c) {
    return map(c);
}

dripper::dripper( Coord source, Coord initVel ) :
    source(source), initVel(initVel), dried(false) {
    //drip_life = one_chance_in(5) ? (one_chance_in(5) ?
	    //6 + rani(20) : 51 + rani(50)) : 200;
    drip_life = 5000;
    max_drips = 20; //6 + rani(30);
}

void dripper::make_drip() {
    drips.push_back(drip(source, initVel));
}

void drip::accelerate(int gravity, int max_vel,
	const Array<int>& height,
	const Array<int>& numdrips)
{
    list<Coord> down_list;
    for (vector<Coord>::iterator dir = orthDirs.begin(); dir != orthDirs.end();
	    dir++)
    {
	if (height(pos) + numdrips(pos)-1
		> height(pos + *dir) +
		numdrips(pos + *dir))
	    down_list.push_back(*dir);
    }
    Coord rand;
    if (down_list.size() == 0
	    || (gravity && one_chance_in(gravity+1)))
    {
	if (gravity)
	    rand = orthDirs[rani(4)];
	else
	    rand = Coord(0,0);
    }
    else
    {
	short unsigned r = down_list.size();
	for (list<Coord>::iterator it = down_list.begin();
		it != down_list.end(); it++)
	    if (one_chance_in(r--))
	    {
		rand = *it;
		break;
	    }
    }

    vel += rand;
    if (vel.sqLen() > max_vel*max_vel) {
	vel *= max_vel;
	vel /= vel.length();
    }
}

void dissolving_cavern::make_rock() {
    // place some random areas of randomly altered hardness
    for (int i = 0; i < hard_areas; i++) {
	const float region_hardness_factor = powf(2,gaussian());
	int region_hardness = int(init_height*region_hardness_factor) ;
	if (region_hardness == 0)
	    region_hardness = 1;

	int dx = max(3, int(10 + 5*gaussian()));
	int dy = max(3, int(10 + 5*gaussian()));
	int sx = rani(w-dx);
	int sy = rani(h-dy);
	for( int x=sx; x<sx+dx; x++ )
	    for( int y=sy; y<sy+dy; y++ )
		height(Coord(x,y)) = region_hardness;
    }

    // place some "cracks"
    for (int i = 0; i < hard_areas/2; i++) {
	Coord s( rani(w), rani(h) );
	Coord e( rani(w), rani(h) );
	for( lineIterator i(s,e); i; i++ )
		height(*i) = max(1, height(*i)/3);
    }


    /*
    // random hardness noise
    for( int x=0; x<w; x++ )
	for( int y=0; y<h; y++ )
	{
	    Coord c(x,y);
	    height(c) = max(3, int(height(c) +
			init_hardness_variation*gaussian()));
	}
    */


    // gently increase hardness towards the map boundaries
    for( int r=0; r<10; r++ ) {
	for( int j=0; j<h; j++ ) {
	    height(Coord(r,j)) += 100-10*r;
	    height(Coord(w-r-1,j)) += 100-10*r;
	}
	for( int i=0; i<w; i++ ) {
	    height(Coord(i,r)) += 100-10*r;
	    height(Coord(i,h-r-1)) += 100-10*r;
	}
    }

    if (encase)
	for (list<dripper>::iterator dripper = drippers.begin();
		dripper != drippers.end(); dripper++)
	{
	    Coord outdir = dirTowards(dripper->initVel);
	    for ( discIterator di(0,encase);
		    di; di++) {
		if (dirTowards(*di) != outdir)
		    height(dripper->source + *di) *= int(1000 / (*di).length());
	    }
	}
}

/*
void dissolving_cavern::make_rock()
{
    hexdir::disc d(GRAD-MAPGEN_BORDER);
    for (hexdir::disc::iterator it = d.begin(); it != d.end(); it++)
    {
	const hexcoord h = hexcoord::centre + *it;
	hardness(h) = init_hardness - init_hardness_variation +
	    rani(1+2*init_hardness_variation);
    }

    // place some random areas of randomly altered hardness
    const dgn_region dgn_bounds_hex =
	dgn_region::perfect_hex(hexcoord::centre, GRAD-(MAPGEN_BORDER+1));
    for (int i = 0; i < hard_areas; i++)
    {
	// random hex
	// see builder_basic_hex() for reasoning
	int dX,dY,dZ,Zp;
	dgn_region myroom;

	dX = 2 + rani(20);

	dY = 2 + rani(20);

	dZ = (rani( dX+dY ) / 2);
	if (dZ < 2)
	    dZ = 2;

	Zp = -rani(dX+dY) - dZ/2;
	do
	{
	    myroom = dgn_region( dgn_bounds_hex.random_point(), dX+1, dY+1,
		    rani(6), Zp, Zp+dZ );
	} while (!dgn_bounds_hex.contains(myroom));

	const int region_hardness_factor = 5+rani(16);
	short unsigned region_hardness =
	    one_chance_in(3) ?
	    init_hardness/region_hardness_factor :
	    init_hardness*region_hardness_factor ;
	if (region_hardness == 0)
	    region_hardness = 1;

	for (dgn_region::contents_iterator it = myroom.begin_contents(); it !=
		myroom.end_contents(); it++)
	    hardness(myroom.pos_in(*it)) = region_hardness;
    }

    if (encase)
	for (vector<dripper>::iterator dripper = drippers.begin();
		dripper != drippers.end(); dripper++)
	{
	    hexdir::disc d2(encase);
	    hexdir outdir = orthDirs[rani(4)];
	    for (hexdir::disc::iterator it = d2.begin(); it != d2.end();
		    it++)
	    {
		if (*it != outdir*((*it).lenDiamond()))
		{
		    const hexcoord h = dripper->source + *it;
		    hardness(h) *= 10;
		}
	    }
	}

    // gently increase hardness towards the map boundaries
    for (int r=0; r<7; r++)
    {
	hexdir::circle c(GRAD-MAPGEN_BORDER-r);
	for (hexdir::circle::iterator it = c.begin(); it != c.end(); it++)
	{
	    const hexcoord h = hexcoord::centre + *it;
	    hardness(h) += 3*(7-r);
	}
    }
}
*/

void dissolving_cavern::dissolve()
{
    // assign begin features
    for (list<dripper>::iterator dripper = drippers.begin();
	    dripper != drippers.end(); dripper++)
    {
	height(dripper->source) = 0;
    }

    // dissolve me a dungeon
    int drip_ticker = 0;
    int time_ticker = 0;
    unsigned int drycount = 0;
    bool dead_drips = false;
    bool diminishdrip = false;

    int debugstep = // set to n>0 to show state every n steps
#ifdef DEBUG
	1000; 
#else
	0;
#endif

    const unsigned int numdrippers = drippers.size();
    while (drycount < numdrippers)
    {
	for (list<dripper>::iterator dripper_it = drippers.begin();
		dripper_it != drippers.end(); dripper_it++)
	{
	    if (dripper_it->dried)
		continue;
	    if (drip_ticker == 0 &&
		dripper_it->drips.size() < dripper_it->max_drips)
		{
		    dripper_it->make_drip();
		    numdrips(dripper_it->source)++;
		}
	    for (vector<drip>::iterator drip_it =
		    dripper_it->drips.begin(); drip_it !=
		    dripper_it->drips.end(); drip_it++)
	    {
		Coord& dpos = drip_it->pos;
		Coord& dvel = drip_it->vel;
		Coord& derr = drip_it->error;
		if (drip_it->dead)
		    continue;
		if (dvel != Coord(0,0))
		{
		    const int N = drip::PRECISION;
		    float vlen = sqrt(dvel.sqLen());
		    Coord verr = (dvel*N)/vlen;
		    Coord toerr = derr + verr;
		    Coord to = dpos;
		    if( toerr.x > N/2 ) {
			toerr.x -= N;
			to.x += 1;
		    }
		    if( toerr.x < N/2 ) {
			toerr.x += N;
			to.x -= 1;
		    }
		    if( toerr.y > N/2 ) {
			toerr.y -= N;
			to.y += 1;
		    }
		    if( toerr.y < N/2 ) {
			toerr.y += N;
			to.y -= 1;
		    }

		    Coord towards = to - dpos;
		    
		    if (height(to) > 0)
		    {
			height(to)--;
			if (height(to) == 0) {
			    // wall eroded
			    height(to) = max(height(dpos),
				    getMinHeight(to));
			}
			else
			    dvel -= towards;
			diminishdrip = true;
		    }
		    else if (in_bounds_strict(to)) {
			if (height(dpos) + numdrips(dpos)-1
				< height(to) + numdrips(to))
			    // going up a slope
			    dvel -= towards;

			numdrips(dpos)--;
			dpos = to;
			derr = toerr;
			numdrips(dpos)++;
		    }
		    else // out of bounds
		    {
			dvel -= towards;
			diminishdrip = true;
		    }
		}

		drip_it->accelerate(gravity, max_vel, height, numdrips);

		if (one_chance_in(dissolve_vert) && height(dpos) > getMinHeight(dpos))
		{
		    height(dpos)--;
		    diminishdrip = true;
		    if ( height(dpos) <= absMinHeight ) {
			dripper_it->dried = true;
			drycount++;
			for (std::vector<drip>::iterator drip_it2 =
				dripper_it->drips.begin(); drip_it2 !=
				dripper_it->drips.end(); drip_it2++)
			    if (!drip_it2->dead)
				numdrips(drip_it2->pos)--;
			break;
		    }
		}
		if (diminishdrip)
		{
		    if (dripper_it->drip_life &&
			    one_chance_in(dripper_it->drip_life))
		    {
			drip_it->dead = true;
			dead_drips = true;
			numdrips(dpos)--;
		    }
		    diminishdrip = false;
		}
	    }
	    if (dead_drips)
	    {
		dripper_it->drips.erase(remove_if(dripper_it->drips.begin(),
			    dripper_it->drips.end(), drip::is_dead),
			dripper_it->drips.end());
		dead_drips = false;
	    }
	}
	drip_ticker = (drip_ticker+1) % drip_rate;
	if( debugstep && (time_ticker % debugstep) == 0 )
	    debugstep = debugView();
	time_ticker++;
    }
}

class DissolvingDebugView : public View {
    int aw, ah;
    const Array<int>& height;
    const Array<int>& numdrips;
    Glyph getGlyph(const Coord& c);
    bool showDepth;
    public:
	DissolvingDebugView(int w, int h, Coord screenTopleft, Coord viewTopleft, const Array<int>& height, const Array<int>& numdrips, int aw, int ah) :
	    View(w, h, screenTopleft, viewTopleft), aw(aw), ah(ah),
	    height(height), numdrips(numdrips),
	    showDepth(false) {}
	void toggleShowDepth() { showDepth = ! showDepth; }
};

Glyph DissolvingDebugView::getGlyph(const Coord& c) {
    char ch = '#';
    int col = GREY;
    if( 0 <= c.x && c.x < aw &&
	    0 <= c.y && c.y < ah ) {
	if( height(c) <= 0)
	    if( numdrips(c) > 0 ) {
		ch = '*';
		col = WHITE;
	    }
	    else
		if( showDepth ) {
		    ch = ( '0' + (-height(c) % 10) );
		    col = ((-height(c) % 20 > 10) ? WHITE : GREY);
		}
		else
		    ch = ' ';
	else if( !showDepth ) {
	    ch = ( '0' + (height(c) % 10) );
	    col = ((height(c) % 20 > 10) ? WHITE : GREY);
	}
    }
    return Glyph(ch, col);
}

int dissolving_cavern::debugView() {
    int vw;
    int vh;
    getmaxyx(stdscr,vh,vw);
    vh--;
    static DissolvingDebugView dv(vw, vh, Coord(0,0), Coord(0,h/2-vh/2), height, numdrips, w, h);
    static Coord centre(w/2,h/2);
    static int step = 1000;

    dv.setFocus(centre);
    dv.drawAll();

    char buf[80];
    snprintf(buf, 80, "dissolving_cavern: "
	    "dis_vert:%d; gravity %d; hardness:%d; "
	    "areas:%d",
	    dissolve_vert, gravity, init_height,
	    hard_areas);
    mvaddstr(vh, 0, buf);
    refresh();


    bool quit = false;
    while( !quit ) {
	char c = getch();
	Coord delta(0,0);
	switch (c) {
	    case 'h': delta = Coord(-vw/2,0); break;
	    case 's': delta = Coord(vw/2,0); break;
	    case 'n': delta = Coord(0,-vh/2); break;
	    case 't': delta = Coord(0,vh/2); break;
	    case 'v': dv.toggleShowDepth(); dv.drawAll(); break;
	    case ' ': quit = true; break;
	    default: {
			 int n = c-'1';
			 if (n < 10) {
			     step = int(pow(10,n));
			     quit = true;
			     break;
			 }
		     }
	}
	if( delta != Coord(0,0)) {
	    centre += delta;
	    dv.setFocus(centre);
	}
    }
    erase();
    return step;
}

/*
void dissolving_cavern::writeDebug(int time)
{
    FILE* log = fopen("dissolveDebug", "a");
    fprintf(log, "%d\n", time);
    for( int j=0; j<h; j++ ) {
	for( int i=0; i<w; i++ ) {
	    Coord c(i,j);
	    fprintf(log, "%d:%d;", height(c), numdrips(c));
	}
	fprintf(log, "\n");
    }
    fclose(log);
}
*/
