#include <curses.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

using namespace std;

#include "config.h"

#include "player.h"
#include "anim.h"
#include "event.h"
#include "coord.h"
#include "fov.h"
#include "view.h"
#include "info.h"
#include "shapes.h"
#include "cave.h"
#include "glyph.h"
#include "messages.h"
#include "random.h"
#include "constants.h"

char _getch() {
    // wrapper around wgetch to handle terminal resizing and refreshing
    int c = wgetch(stdscr);
    switch(c) {
	case KEY_RESIZE:
	    setGeometry(); return 0;
	case 12: // ^L
	    redrawwin(stdscr); refresh(); return 0;
	default:
	    return c;
    }
}
    
Player::Player(Coord c, Glyph glyph) : Actor(c, glyph) {
    isFocus = true;
    activeBranches.insert(0);
    setBranchStati();
    maxhp = 600;
    hp = 400;
    energy = maxenergy = 6;
    fuel = 0;
    autofiring = false;
    examining = false;
    for (int i=0; i<INIT_LAG; i++)
	branches[0].push_back(Order(MoveOrder,Coord(0,0)));
    lag = INIT_LAG;
}

bool Player::tryMove(Coord dir, bool force) {
    if (dir == Coord(0,0))
	return true;
    Coord newpos = getPos() + dir;
    if( cave.getTile(newpos) == TILE_SURFACE ) {
	static bool warnedSurface = false;
	if (!warnedSurface)
	    logroll.add("You can't risk discovery by sending the probe back out there.");
	warnedSurface = true;
	return false;
    }
    else if ( cave.getTile(newpos) == TILE_EXIT ) {
	logroll.add("There's nothing up there but dust and wind.");
	return false;
    }
    else if( cave.getTile(newpos) == TILE_LAVA ) {
	logroll.add("The probe melts to slag, leaving nothing traceable to Earth. Congratulations!");
	hp = 0;
	return true;
    }

    bool success = Actor::tryMove(dir, force);
    if (success) {
	Objects os = cave.objectsInCell(getPos());
	for (Objects::iterator o = os.begin(); o!=os.end(); o++)
	    if ((*o)->collectable()) {
		switch ((*o)->collectable()) {
		    case COLLECTABLE_FUEL: fuel++; break;
		    default: ;
		}
		cave.deleteObject(*o);
	    }
    }
    return success;
}

bool Player::clearPathInDir(Coord dir) {
    static const int dist=4;

    AnimFrame moveframe;
    bool blocked = false;
    for (int i=0; i < dist; i++) {
	const Coord from = getPos()+dir*i;
	const Coord to = from+dir;
	Glyph g = gameView.getGlyph(to);
	if (cave.canPass(from, dir, false, flying))
	    g.col = B_GREEN;
	else {
	    g.col = B_RED;
	    blocked = true;
	}
	moveframe.push_back(FrameSquare(to, g));
	if (blocked)
	    break;
    }
    anim.push_back(moveframe);
    if (blocked)
	return false;

    const Coord dest = getPos() + dir*dist;
    AnimFrame nastyframe;
    bool foundNasty = false;
    for ( rectIterator ri(dest - Coord(dist,dist), dest + Coord(dist,dist));
	    ri; ri++) {
	Objects os = cave.objectsInCell(*ri);
	if (os.empty() || !inFOV(*ri))
	    continue;
	for (Objects::iterator o = os.begin(); o != os.end(); o++)
	    if ((*o)->isThreat()) {
		nastyframe.push_back( FrameSquare(*ri, Glyph('!', B_RED)) );
		foundNasty = true;
	    }
    }
    if (foundNasty) {
	anim.push_back(nastyframe);
	return false;
    }
    return true;
}


ActionReturn Player::interact() {
    Events events;
    events.push_back(new InteractEvent(1, this));

    dirtyBranches();
    setBranchStati();
    dirtyBranches();

    gameView.setFocus(getPos());
    anim.draw();
    updateSeenObjects();
    dirtyFov();
    gameView.drawDirty();
    infoPanel.draw();
    helpPanel.draw();
    logroll.draw();
    indicators.draw();

    if (hp <= 0) {
	char c = _getch();
	if (c == 'q') {
	    endwin();
	    exit(0);
	}
	static int fovShrinkCount = 0;
	fovShrinkCount = (fovShrinkCount + 1) % 3;
	if (!fovShrinkCount && fovDist) {
	    dirtyFov();
	    fovDist--;
	    recompute_fov = true;
	}
	return ActionReturn(events);
    }

    bool consumedOrder = false;
    do {
	Orders::iterator i = branches[0].begin();
	Order order = *i;
	if (order.instant) {
	    branches[0].erase(i);
	    if (order.type == ToggleAutofireOrder)
		autofiring = !autofiring;
	}
	else {
	    if (autofiring && autofire()) {
		logroll.add("Autofire! [space]");
		char c;
		do c = _getch(); while (c!=' ');
	    }
	    else if (order.type == RunOrder && energy < RUN_COST) {
		logroll.add("Run: insufficient charge!");
		tryMove(order.dir);
		i->type = MoveOrder;
	    }
	    else {
		consumedOrder = true;

		appendOrders(); // XXX: may invalidate i

		branches[0].erase(branches[0].begin());

		switch (order.type) {
		    case MoveOrder:
			tryMove(order.dir);
			break;
		    case RunOrder:
			{
			    assert (energy >= RUN_COST);
			    energy -= RUN_COST;

			    Coord frompos = getPos();
			    for (int i=0; i<2; i++) {
				Coord topos = frompos + order.dir;
				bool jumping = (cave.getTileDef(frompos).cliff == 1
					&& cave.getTileDef(topos).cliff == -1
					&& cave.cellPassable(topos));
				bool moved = tryMove(order.dir, jumping);
				if (moved && i==0) {
				    AnimFrame frame;
				    updateSeenObjects();
				    frame.push_back( FrameSquare(frompos) );
				    frame.push_back( FrameSquare(getPos()) );
				    anim.push_back(frame);
				}
				if (jumping)
				    hit(2, "The probe is damaged by the fall.");
				frompos = getPos();
			    }
			}
			break;
		    case FireOrder:
			fire(order.dir);
			break;
		    case ToggleAutofireOrder:
			autofiring = !autofiring;
			break;
		    case AutofireOrder:
			if (energy < AUTOFIRE_COST + FIRE_COST) {
			    logroll.add("Autofire: insufficient charge!");
			    break;
			}
			energy -= AUTOFIRE_COST;
			autofire();
			break;
		    case TestClearOrder:
			if (energy < TEST_CLEAR_COST) {
			    logroll.add("Check path: insufficient charge!");
			    break;
			}
			energy -= TEST_CLEAR_COST;
			if (clearPathInDir(order.dir)) {
			    logroll.add("Path clear.");
			    removeBranch(order.onFail);
			}
			else {
			    logroll.add("Path not clear.");
			    removeBranch(0);
			    moveBranch(order.onFail, 0);
			}
			break;
		    case ConsumeFuelOrder:
			if (fuel>0) {
			    fuel--;
			    energy=maxenergy;
			}
			else
			    logroll.add("Recharge: out of fuel!");
		}
	    }
	    break;
	}
    } while (!consumedOrder);

    if (autofiring)
	energy--;

    energy++;

    assert (energy>=0);
    if (energy > maxenergy) {
	energy = maxenergy;
	if (hp < maxhp && hp > 0)
	    hp = min(maxhp, hp + 1 + ( maxhp - hp )/100);
    }

    return ActionReturn(events);
}

Coord dirOfChar(char c) {
    switch (c) {
#ifdef DVORAK
	case 'h': case '4': return Coord(-1,0);
	case 't': case '2': return Coord(0,1);
	case 'n': case '8': return Coord(0,-1);
	case 's': case '6': return Coord(1,0);
	case 'g': case '7': return Coord(-1,-1);
	case 'c': case '9': return Coord(1,-1);
	case 'b': case '1': return Coord(-1,1);
	case 'm': case '3': return Coord(1,1);
#else
	case 'h': case '4': return Coord(-1,0);
	case 'j': case '2': return Coord(0,1);
	case 'k': case '8': return Coord(0,-1);
	case 'l': case '6': return Coord(1,0);
	case 'y': case '7': return Coord(-1,-1);
	case 'u': case '9': return Coord(1,-1);
	case 'b': case '1': return Coord(-1,1);
	case 'n': case '3': return Coord(1,1);
#endif
	case 27: { // ESC - check for cursor keys
		     if (_getch() == '[') {
			 char c2 = _getch();
			 switch (c2) {
			     case 'A': return Coord(0,-1);
			     case 'B': return Coord(0,1);
			     case 'C': return Coord(1,0);
			     case 'D': return Coord(-1,0);
			     case '1': case '4': case '5': case '6':
				       if (_getch() == '~')
					   switch (c2) {
					       case '1': return Coord(-1,-1);
					       case '5': return Coord(1,-1);
					       case '6': return Coord(1,1);
					       case '4': return Coord(-1,1);
					       default:;
					   }
			     default :;
			 }
		     }
		 } // fallthrough
	default: return Coord(0,0);
    }
}

int Player::nextFreeBranch(int branch) {
    for (branch++; branch < MAX_BRANCHES; branch++)
	if (activeBranches.find(branch) == activeBranches.end())
	    return branch;
    return -1;
}

void Player::examineMode(bool largeScale) {
    examining = true;
    examineCell = getPos();

    logroll.clearBox();
    Messages examineMessages;
    int w,h;
    getmaxyx(stdscr,h,w);
    examineMessages.setGeom(firstMessageLine(h),w,1);

    gameView.setFocus(examineCell);
    gameView.setDirty(examineCell);
    gameView.drawDirty();

    const int dx = largeScale ? 3*gameView.w/7 : 1;
    const int dy = largeScale ? 3*gameView.h/7 : 1;

    while (true) {
	Coord delta = dirOfChar(_getch());
	delta.x *= dx;
	delta.y *= dy;
	if( delta != Coord(0,0)) {
	    gameView.setDirty(examineCell);
	    examineCell += delta;
	    gameView.setFocus(examineCell);
	    gameView.setDirty(examineCell);
	    gameView.drawDirty();

	    const char* description="";
	    Objects os = seenObjects.objectsInCell(examineCell);
	    for (Objects::iterator o = os.begin(); o != os.end(); o++) {
		description = (*o)->describe();
		if (strlen(description) > 0)
		    break;
	    }
	    if (strlen(description) == 0 && cave.isExplored(examineCell))
		description = cave.getTileDef(examineCell).description;
	    if (strlen(description) > 0)
		examineMessages.add(description);

	    examineMessages.draw();
	}
	else
	    break;
    }
    logroll.redraw();
    examining = false;
    gameView.setDirty(examineCell);
    gameView.drawDirty();
    gameView.setFocus(getPos());
}

Order Player::getOrder(int branch) {
    int branchCol = branchGlyph(branch).col;
    while (true) {
	char c = _getch();
	Coord dir(0,0);
	switch (c) {
	    case ' ': case '5': return Order(MoveOrder, Coord(0,0));
	    case 'p':
	      {
		  dir = dirOfChar(_getch());
		  if (dir == Coord(0,0)) continue;
		  int forkBranch = nextFreeBranch(branch);
		  if (branchStati[branch].energy < TEST_CLEAR_COST)
		      logroll.add("Check path: insufficient charge!", branchCol);
		  else if (forkBranch == -1)
		      logroll.add("Out of memory!", branchCol);
		  else {
		      activeBranches.insert(forkBranch);
		      return Order(TestClearOrder, dir, forkBranch);
		  }
		  break;
	      }
	    case 'f':
	      dir = dirOfChar(_getch());
	      if (dir == Coord(0,0)) continue;
	      if (branchStati[branch].energy < FIRE_COST)
		  logroll.add("Fire: insufficient charge!", branchCol);
	      else
		  return Order(FireOrder, dir);
	      break;
	    //case 'a': return Order(ToggleAutofireOrder, Coord(0,0), -1, true);
	    case 'a':
	      if (branchStati[branch].energy < AUTOFIRE_COST + FIRE_COST)
		  logroll.add("Autofire: insufficient charge!", branchCol);
	      else
		  return Order(AutofireOrder, Coord(0,0));
	      break;
	    case 'r':
	      dir = dirOfChar(_getch());
	      if (dir == Coord(0,0)) continue;
	      if (branchStati[branch].energy < RUN_COST)
		  logroll.add("Run: insufficient charge!", branchCol);
	      else
		  return Order(RunOrder, dir);
	      break;
#ifndef DVORAK
	    case 'c':
#else
	    case 'e':
#endif
	    case '*':
	      if (fuel > 0)
		  return Order(ConsumeFuelOrder);
	      else
		  logroll.add("Recharge: out of fuel!");
	      break;
	    case 'x':
	      examineMode();
	      break;
	    case 'X':
	      examineMode(true);
	      break;
	    case 'v':
	      setBranchStati(true);
	      anim.draw();
	      break;
#ifdef DEBUG
	    case '&':
	      for ( rectIterator ri(getPos() - Coord(500, 500),
			  getPos() + Coord(500, 500)); ri; ri++ )
		  cave.setExplored(*ri);
	      break;
	    case '^':
	      increaseLag();
	      break;
#endif
	    case 'q':
	      logroll.add("Really quit?");
	      logroll.draw();
	      c = _getch();
	      if (c == 'y') {
		  endwin();
		  exit(0);
	      }
	      break;
	    case '?':
	      logroll.add("hjklyubn/42867913 [f]ire [r]un [a]utofire [p]ath re[c]harge e[x]amine [v]iew lag");
	      break;
	    default:
	      dir = dirOfChar(c);
	      if (dir != Coord(0,0))
		  return Order(MoveOrder, dir);
	}
	logroll.draw();
    }
}

const char* Player::describe() const {
    return "The hapless probe left to explore this world.";
}

const Pattern firePattern(
	"     \n"
	"  ## \n"
	"*####\n"
	"  ## \n"
	"     \n",

	"     \n"
	"  ## \n"
	" ### \n"
	" ##  \n"
	"*    \n");

AnimFrame fakeFire(Coord pos, Coord dir) {
    AnimFrame frame(true);
    vector<Coord> cs = firePattern.place(pos,dir);
    for (vector<Coord>::iterator c=cs.begin(); c!=cs.end(); c++)
	frame.push_back( FrameSquare(*c, Glyph('#', BLUE)) );
    return frame;
}
void Player::fire(Coord dir) {
    if (energy < FIRE_COST) {
	logroll.add("Fire: insufficient charge!");
	return;
    }
    energy -= FIRE_COST;

    AnimFrame frame;
    vector<Coord> cs = firePattern.place(getPos(),dir);
    for (vector<Coord>::iterator c=cs.begin(); c!=cs.end(); c++) {
	if (inFOV(*c)) {
	    frame.push_back( FrameSquare(*c, Glyph('#', CYAN)) );
	    Objects os = cave.objectsInCell(*c);
	    for (Objects::iterator o = os.begin(); o != os.end(); o++)
		(*o)->hit();
	}
    }
    anim.push_back(frame);
}
bool Player::autofire() {
    if (energy < FIRE_COST)
	return false;
    AnimFrame frame;
    Coord fdir(0,0);
    int closest=1000;
    for (vector<Coord>::iterator dir = dirs.begin(); dir != dirs.end(); dir++) {
	frame.push_back(FrameSquare(getPos()+*dir, Glyph('?',CYAN)));
	vector<Coord> cs = firePattern.place(getPos(),*dir);
	for (vector<Coord>::iterator c=cs.begin(); c!=cs.end(); c++) {
	    int sqdist = (*c-getPos()).sqLen();
	    if (sqdist < closest && inFOV(*c) ) {
		Objects os = cave.objectsInCell(*c);
		for (Objects::iterator o = os.begin(); o != os.end(); o++)
		    if ((*o)->isThreat()) {
			fdir = *dir;
			closest = sqdist;
			frame.push_back(FrameSquare(getPos()+*dir,
				    Glyph('!',CYAN)));
		    }
	    }
	}
    }
    anim.push_back(frame);
    if (fdir != Coord(0,0)) {
	fire(fdir);
	return true;
    }
    return false;
}


Glyph Player::branchGlyph(int branch) {
    Glyph g = glyph;
    if (branchStati[branch].animTime > 0)
	g.ch = '0' + branchStati[branch].animTime%10;
    g.col = (branch == selectedBranch ? B_RED : RED ) + (branch%6);
    return g;
}
Glyph Player::getGlyph() const {
    if (cave.getTile(getPos()) == TILE_LAVA)
	return tileDefOfTile(TILE_LAVA).glyph;
    Glyph g = glyph;
    if (hp <= 0)
	g.ch = '%';
    g.col = (selectedBranch == -1 ? WHITE : GREY );
    return g;
}

void Player::appendOrders() {
    set<int> iterBranches = activeBranches;
    assert (!iterBranches.empty());
    gameView.setDirty(getPos());
    for (set<int>::iterator br=iterBranches.begin(); br!=iterBranches.end(); br++) {
	selectedBranch = *br;

	Order order;
	do {
	    gameView.setFocus(branchStati[*br].pos, true);
	    dirtyBranch(*br);
	    gameView.drawDirty();
	    infoPanel.draw();

	    order = getOrder(*br);
	    branches[*br].push_back(order);
	    dirtyBranch(*br);
	    branchStati[*br].update(order);
	    dirtyBranch(*br);
	    if (order.onFail > -1)
		branchStati[order.onFail] = branchStati[*br];
	} while (order.instant);
    }
    selectedBranch = -1;
}

void Player::dirtyBranch(int branch) {
    gameView.setDirty(branchStati[branch].pos);
}
void Player::dirtyBranches() {
    for (set<int>::iterator br=activeBranches.begin(); br!=activeBranches.end(); br++)
	dirtyBranch(*br);
}

void Player::removeBranch(int branch) {
    dirtyBranch(branch);
    for (Orders::iterator i = branches[branch].begin(); i != branches[branch].end(); i++)
	if (i->onFail > -1)
	    removeBranch(i->onFail);
    activeBranches.erase(branch);
    branches[branch].clear();
}
void Player::moveBranch(int from, int to) {
    branches[to] = branches[from];
    activeBranches.erase(from);
    branches[from].clear();
    activeBranches.insert(to);
}

int Player::branchPosInCell(Coord pos) {
    int b = -1;
    for (set<int>::iterator br=activeBranches.begin(); br!=activeBranches.end(); br++)
	if (branchStati[*br].pos == pos) {
	    if (*br == selectedBranch)
		return *br;
	    b = *br;
	}
    return b;
}

BranchStatus::BranchStatus(Player* pl) : pos(pl->getPos()), energy(pl->energy),
    maxenergy(pl->maxenergy), fuel(pl->fuel), autofiring(pl->autofiring),
    animTime(0) {}

void BranchStatus::update(const Order& order, bool animate) {
    AnimFrame frame;
    switch (order.type) {
	case MoveOrder:
	case RunOrder:
	    {
		int d = 1;
		if (order.type == RunOrder && energy >= RUN_COST) {
		    energy -= RUN_COST;
		    d = 2;
		}
		for (int i=0; i<d; i++) {
		    Coord newpos = pos + order.dir;
		    if ( ( !cave.isExplored(newpos) || cave.canPass(pos,
				    order.dir, true, player.flying)
				|| ( d==2 && cave.getTileDef(pos).cliff == 1
				    && cave.getTileDef(newpos).cliff == -1
				    && cave.cellPassable(newpos) ) ) &&
			    ( ! ( player.inFOV(newpos) &&
				  cave.contains_if(newpos,
				      &isStaticObstruction) ) ) ) {
			const Coord oldpos = pos;

			pos = newpos;

			if (animate) {
			    frame.push_back(FrameSquare(oldpos));
			    frame.push_back(FrameSquare(pos));
			}
		    }
		}
	    }
	    break;
	case FireOrder:
	    if (energy >= FIRE_COST) {
		energy -= FIRE_COST;
		if (animate)
		    frame = fakeFire(pos, order.dir);
	    }
	    break;
	case AutofireOrder:
	    energy -= AUTOFIRE_COST;
	    if (animate) {
		frame.cleanAfter = true;
		for (int d=0; d<8; d++)
		    frame.push_back(FrameSquare(pos+dirs[d],
				Glyph('?',BLUE)));
	    }
	    break;
	case TestClearOrder:
	    energy -= TEST_CLEAR_COST;
	    if (animate) {
		frame.cleanAfter = true;
		frame.push_back(FrameSquare(pos+order.dir,
			    Glyph('?',player.branchGlyph(order.onFail).col)));
	    }
	    break;
	case ConsumeFuelOrder:
	    if (fuel>0) {
		fuel--;
		energy=maxenergy;
	    }
	    break;
	case ToggleAutofireOrder:
	    autofiring = !autofiring; break;
	default : ;
    }
    if (!order.instant) {
	if (autofiring)
	    energy--;
	energy++;
	if (energy > player.maxenergy)
	    energy = player.maxenergy;
    }
    if (animate) {
	frame.push_back(FrameSquare(pos));
	anim.push_back(frame);
    }
}

void Player::_setBranchStati(const BranchStatus& status, int branch,
	bool animate) {
    branchStati[branch] = status;
    for (Orders::iterator i = branches[branch].begin(); i !=
	    branches[branch].end(); i++) {
	if (animate) {
	    selectedBranch = branch;
	    branchStati[branch].animTime++;
	}
	branchStati[branch].update(*i, animate);
	if (i->onFail > -1)
	    _setBranchStati(branchStati[branch], i->onFail, animate);
    }
    if (animate)
	branchStati[branch].animTime = 0;
}
void Player::setBranchStati(bool animate) {
    if (animate) {
	AnimFrame frame;
	for (set<int>::iterator br=activeBranches.begin(); br!=activeBranches.end(); br++) {
	    Coord oldpos = branchStati[*br].pos;
	    branchStati[*br] = BranchStatus(this);
	    frame.push_back( FrameSquare(oldpos) );
	}
	frame.push_back( FrameSquare(getPos()) );
	anim.push_back(frame);
    }

    int realSelected = selectedBranch;
    _setBranchStati(BranchStatus(this), 0, animate);
    selectedBranch = realSelected;
}

void Player::hit(int damage, const char* message) {
    AnimFrame frame;

    hp -= 100*damage;
    logroll.add(message);
    frame.push_back( FrameSquare(getPos(), Glyph('x', WHITE)) );
    if (hp <= 0) {
	logroll.add("Probe destroyed. The wreckage could be traced to Earth. You have failed.");
	gameView.setDirty(getPos());
	while (!activeBranches.empty())
	    removeBranch(*activeBranches.begin());
    }
    anim.push_back(frame);
}

void Player::increaseLag() {
    lag++;
    if (hp > 0) {
	logroll.add(lag == 2 ?
		"The communications lag worsens as your ship breaks out of orbit." :
		lag == 5 ?
		"Telemetry lag increased, as one of this planet's moons glides past your window." :
		lag == 10 ?
		"Telemetry lag increased; this rock is now just a distant speck." :
		lag == 20 ?
		"Telemetry lag increased; the ship is now well on its way to the next world." :
		"Telemetry lag increased.");
	logroll.draw();

	appendOrders();
    }
}

bool Player::setGoal(Coord* goal) {
    static int SENSOR_RANGE=300;
    for ( rectIterator ri(Coord(-SENSOR_RANGE, -SENSOR_RANGE), Coord(SENSOR_RANGE,SENSOR_RANGE)); ri; ri++ )
	if (ri->sqLen() <= SENSOR_RANGE*SENSOR_RANGE &&
		cave.getTile(getPos() + *ri) == TILE_LAVA) {
	    *goal = getPos() + *ri;
	    return true;
	}
    return false;
}

void Player::updateSeenObjects() {
    // XXX: we loop over FOV every turn, updating our current knowledge of
    // objects accordingly. This is pretty wasteful, and probably ought to be
    // replaced with something along the lines of setDirty.
    for ( rectIterator ri(getPos() - Coord(fovDist, fovDist),
		getPos() + Coord(fovDist, fovDist)); ri; ri++ )
	if (inFOV(*ri)) {
	    seenObjects.erase(*ri);
	    Objects os = cave.objectsInCell(*ri);
	    for (Objects::iterator i = os.begin(); i!=os.end(); i++)
		seenObjects.insert((*i)->clone());
	}
}

ActionReturn LagIncreaseEvent::execute() {
    player.increaseLag();
    return ActionReturn(Events(1, new LagIncreaseEvent(LAG_INC_TIME)));
}
