#include <assert.h>
using namespace std;

#include "shapes.h"
#include "coord.h"

lineIterator Line::line() const {
    return lineIterator(s,e);
}
bool Line::contains(Coord c) const {
    for( lineIterator i=line(); i; i++ )
	if( c == *i )
	    return true;
    return false;
}

// Line::intersectsLine : returns Bresenham distance along line that 'other'
// intersects this line, or else -1.
int Line::intersectsLine(Line other) const {
    const Coord a = other.s - s;
    const Coord d = e-s;
    const Coord od = other.e-other.s;
    const int l = od.x * d.y - od.y * d.x;

    if( l == 0 ) // parallel lines
	return -1;

    // Euclidean dist along 'other' of intersection point:
    const float t = (d.y*a.x - d.x*a.y)/l;
    if( t < 0 ) // before start of other line
	return -1;
    const Coord oc(int(t*od.x), int(t*od.y));
    if( Line(Coord(0,0),oc).len() > other.len() ) // after end of other line
	return -1;
    const Coord c = other.s + oc - s;
    if( d.x*(c.y*d.x - c.x*d.y) < 0 ) // before start of this line
	return -1;
    const int dist = Line(Coord(0,0),c).len();
    if( dist > len() ) // after end of this line
	return -1;
    return dist;
}

// Line::dist : returns Bresenham distance along line of a normal dropped from
// c
int Line::dist(Coord c) const {
    return 0;
}

Line Line::normalTo(Coord c) const {
    return Line((*this)[dist(c)],c);
}

Coord Line::operator[](int n) const {
    lineIterator i=line();
    for( int j=0; j<n; i++,j++ )
	assert(i);
    return *i;
}

// Line::len : returns Bresenham length, so we have: [len()] == e.
int Line::len() const {
    return max(abs(e.x-s.x), abs(e.y-s.y));
}

lineIterator::operator bool() const {
    return (d <= dist);
}

lineIterator& lineIterator::operator++() {
    c+=mainDir;
    d++;
    error += slope;
    if (error > dist/2) {
	c+=offDir;
	error -= dist;
    }
    return *this;
}
lineIterator& lineIterator::operator--() {
    c-=mainDir;
    d--;
    error -= slope;
    if (error < -dist/2) {
	c-=offDir;
	error += dist;
    }
    return *this;
}

lineIterator lineIterator::operator++(int) {
    lineIterator tmp(*this);
    ++(*this);
    return(tmp);
}
lineIterator lineIterator::operator--(int) {
    lineIterator tmp(*this);
    --(*this);
    return(tmp);
}

const Coord& lineIterator::operator*() const {
    return c;
}

lineIterator::lineIterator(Coord s, Coord e) : c(s), d(0), error(0) {
    const int dx = e.x - s.x;
    const int dy = e.y - s.y;

    if( abs(dx) > abs(dy) ) {
	mainDir = Coord( dx > 0 ? 1 : -1, 0 );
	offDir = Coord(0, dy > 0 ? 1 : -1 );
	slope = abs(dy);
	dist = abs(dx);
    }
    else {
	mainDir = Coord(0, dy > 0 ? 1 : -1 );
	offDir = Coord( dx > 0 ? 1 : -1, 0 );
	slope = abs(dx);
	dist = abs(dy);
    }
}

rectIterator::operator bool() const {
    return ( c.y <= e.y );
}
Coord rectIterator::operator*() const {
    return c;
}
const Coord* rectIterator::operator->() const {
    return &c;
}
rectIterator rectIterator::operator++() {
    if( c.x < e.x )
	c.x++;
    else {
	c.x = s.x;
	c.y++;
    }
    return (*this);
}
rectIterator rectIterator::operator++(int) {
    rectIterator tmp(*this);
    ++(*this);
    return tmp;
}

discIterator discIterator::operator++() {
    do { ri++; } while (ri && (*ri-centre).sqLen() > rsq);
    return (*this);
}

discIterator discIterator::operator++(int) {
    discIterator tmp(*this);
    ++(*this);
    return tmp;
}

vector<Coord> patOfStr(const char* pat) {
    vector<Coord> cs;
    Coord pivot;
    int x=0,y=0;
    for (; *pat != '\0'; pat++) {
	switch (*pat) {
	    case '#' : cs.push_back(Coord(x,y)); break;
	    case '*' : pivot = Coord(x,y); break;
	    case '\n': y++; x=0; continue;
	    default : ;
	}
	x++;
    }
    for (vector<Coord>::iterator c = cs.begin(); c != cs.end(); c++)
	*c -= pivot;
    return cs;
}

Pattern::Pattern(const char* epat, const char* nepat) {
    east = patOfStr(epat);
    northeast = patOfStr(nepat);
}

vector<Coord> Pattern::place(Coord origin, Coord dir) const {
    int i = dirIndex(dir);
    int rot = i/2;
    const vector<Coord>* pat = i%2 ? &northeast : &east;

    vector<Coord> ret;
    for (vector<Coord>::const_iterator c=pat->begin(); c != pat->end(); c++)
	ret.push_back(origin + c->rotated(rot));
    return ret;
}
